////////////////////////////////////////////////////////////////////////////////////////////
// ImViewW32.h - Image display class for the win32 framework
// Based on VisImageWnd.h Copyright  1997-2000 Microsoft Corporation, All Rights Reserved
// Adapted by Alexandre Bernardino - IST, 2002
////////////////////////////////////////////////////////////////////////////////////////////

#ifndef IMVIEWW32_H
#define IMVIEWW32_H

#define WM_FORCEPAINT (WM_USER+1)
#define MAX_WINDOW_NAME 20

#include <windows.h>
#include <assert.h>

//forward class decaration
class CImViewW32;

class CImViewW32Manager
{
public:
	// UNDONE:  There should be a static CImViewW32Manager object in some CPP file.
	// It's ctor should initialize a critical section.  CImViewW32 ctors should call
	// the add method and CImViewW32 dtors should call the remove method.  There
	// should be methods to get a table of (duplicate) thread handles for all CImViewW32
	// objects in use.  These tables can be used to wait for or kill all CImViewW32 threads.
	// The CImViewW32Manager dtor should kill all CImViewW32 threads and destroy
	// the critical section before exiting.
	CImViewW32Manager(void);
	~CImViewW32Manager(void);

	void AddImViewW32(CImViewW32 *pimagewnd);
	void RemoveImViewW32(CImViewW32 *pimagewnd);

private:
	// UNDONE:  Critical section and list entries
	// (Keep a list of CImViewW32 in use, and provide methods
	// to gett duplicate handles for all threads in use.)
};

class CImViewW32
{
private:
	char m_DIBHeader[ sizeof(BITMAPINFOHEADER) + sizeof(RGBQUAD)*256 ];  // Header for DIB GRAY images
	int m_iWidth, m_iHeight, m_iBitcount, m_iChannels;
	byte *m_pbData; 
	int m_iFlip, m_iStride;
	char m_szWindowName[MAX_WINDOW_NAME];
public:
	// UNDONE:  Provide a default constructor and public create methods
	// that can accept an image reference and a title string.
	CImViewW32(char *name, int width, int height, int bitcount)
		: m_hEventInitialized(0),
		m_hThread(0),
		m_dwThreadID(0),
		m_hwnd(0),
		m_pbData(0),
		m_fKillThread(false),
		m_fThreadRunning(false)
	{
		/*if ((RegisterClassIfNeeded() != 0)
				&& (refimage.IsValid()))*/
		if(RegisterClassIfNeeded() != 0)
		{
			assert(width >= 0);
			assert((width % 4) == 0);
			assert(height != 0);
			assert((bitcount == 8) || (bitcount == 24) || (bitcount == 32));

			strncpy( m_szWindowName, name, MAX_WINDOW_NAME );

			m_iWidth = width;
			m_iHeight = height;
			m_iBitcount = bitcount;
			m_iChannels = bitcount/8;
				

			// CREATING BITMAP HEADER
			BITMAPINFOHEADER* dibhdr = (BITMAPINFOHEADER*)m_DIBHeader;
			dibhdr->biSize = sizeof( BITMAPINFOHEADER );
			dibhdr->biPlanes = 1;			
			dibhdr->biCompression = BI_RGB;		
			dibhdr->biXPelsPerMeter = 0;
			dibhdr->biYPelsPerMeter = 0;
			dibhdr->biClrUsed = 0;
			dibhdr->biClrImportant = 0; 	
			dibhdr->biWidth = (((m_iWidth*m_iChannels - 1)/8)+1)*8/m_iChannels; // to cope with 8 byte aligned image data
			dibhdr->biHeight = m_iHeight;
			dibhdr->biBitCount = m_iBitcount;
			dibhdr->biSizeImage = 0;
			

			// CREATING GRAY COLOR TABLE IF NEEDED
			if(m_iBitcount == 8)
			{
				COLORREF* rgb = (COLORREF*)( m_DIBHeader + sizeof(BITMAPINFOHEADER) );
				for( int i = 0; i < 256; i++) rgb[i] = RGB( i,i,i );
			}


			//m_pimage = refimage.Clone();
			m_hEventInitialized = CreateEvent(0, true, false, 0);
			
			/*unsigned int uiThreadID;
			m_hThread = (HANDLE) _beginthreadex(0, 0, StaticThreadMain, this, 0, &uiThreadID);*/

			unsigned long uiThreadID;
			m_hThread = (HANDLE) CreateThread(0, 0, StaticThreadMain2, this, 0, &uiThreadID );

			m_dwThreadID = uiThreadID;

			if (m_hThread != 0)
			{
				WaitForSingleObject(m_hEventInitialized, INFINITE);
				SetThreadPriority(m_hThread, THREAD_PRIORITY_BELOW_NORMAL);
			}

			CloseHandle(m_hEventInitialized);
			m_hEventInitialized = 0;
		}
	};

	~CImViewW32(void)
	{
		KillThread();
		if (m_hThread != 0)
		{
			CloseHandle(m_hThread);
			m_hThread = 0;
		}

/*		if (m_pimage != 0)
		{
			delete m_pimage;
			m_pimage = 0;
		}*/
	};

	BOOL SetPos(int x, int y )
	{
		if(m_hwnd)
			return SetWindowPos(m_hwnd, HWND_TOP, x, y, 0, 0, SWP_NOSIZE | SWP_SHOWWINDOW);
		else
			return FALSE;
	};

	bool HasWindow(void) const
	{
		return (m_hwnd != 0);
	};
	bool IsThreadRunning(void) const
	{
		return m_fThreadRunning;
	};
	
	HWND Hwnd(void) const
	{
		return m_hwnd;
	};

	// This will not attempt to kill the thread, but it will return
	// before cmsTimeout milliseconds if the thread is killed by a user
	// action (closing the window).
	void WaitForThreadToQuit(DWORD cmsTimeout = INFINITE)
	{
		WaitForSingleObject(m_hThread, cmsTimeout);
	};


	// This will attempt to kill the thread, but it will return after
	// cmsTimeout milliseconds even if the thread has not been killed.
	void KillThread(DWORD cmsTimeout = INFINITE)
	{
		// This will attempt to kill the thread, but it will return after
		// cmsTimeout milliseconds even if the thread has not been killed.
		if (m_hThread != 0)
		{
			// Kill it.
			::PostThreadMessage(m_dwThreadID, WM_QUIT, 0, 0);
			m_fKillThread = true;

			if (cmsTimeout != 0)
				WaitForThreadToQuit(cmsTimeout);
		}
	}
	
	// This will redraw the image
	void Refresh()
	{
		if(!m_fKillThread)
			::PostThreadMessage(m_dwThreadID, WM_PAINT, 0, 0);
		/*if(m_hwnd)
			::PostMessage(m_hwnd,WM_PAINT,0,0);*/
		/*if(m_hwnd)
			::RedrawWindow(m_hwnd,0,0,RDW_INTERNALPAINT);*/
	}

	void SetData(byte *pbData, int stride)
	{
		if(stride < 0)
		{
			m_iStride = -stride;
			m_iFlip = 1;
		}
		else
		{
			m_iStride = stride;
			m_iFlip = -1;
		}
		m_pbData = pbData;
	}

	// LATER:  Add a static table to keep track of these threads, then
	// add methods to wait for all threads to quit and to kill all threads.


	// LATER:  Add an auto-delete flag.

	// LATER:  Remember the time of window creation and add options to destroy
	// the window after a fixed amount of time or to keep only the last N
	// windows open?

	// LATER:  Window position option (or better automatic positioning)?

	// LATER:  Flag to specify whether window should be activated when it is
	// displayed?

protected:

	static ATOM RegisterClassIfNeeded(void)
	{
		static bool fRegisteredClass = false;
		static ATOM atomRet;

		if (!fRegisteredClass)
		{
			s_szAppName[0] = 0;
			GetModuleFileName(0, s_szAppName, sizeof(s_szAppName));

			WNDCLASSEX wcex; 
			wcex.style         = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
			wcex.lpfnWndProc   = (WNDPROC) CImViewW32::WndProc;
			wcex.cbClsExtra    = 0;
			wcex.cbWndExtra    = 0;
			wcex.hInstance     = ::GetModuleHandle(0);
			// LATER:  wcex.hIcon         = LoadIcon (hInstance, szAppName);
			wcex.hIcon         = 0;
			wcex.hCursor       = LoadCursor(0, IDC_ARROW);
			wcex.hbrBackground = (HBRUSH) (COLOR_WINDOW+1);
			wcex.lpszMenuName  = "WIN95";

			wcex.lpszClassName = s_szAppName; 

			// Added elements for Windows 95: 
			wcex.cbSize = sizeof(WNDCLASSEX); 
			// LATER:  wcex.hIconSm = LoadIcon(wcex.hInstance, "SMALL"); 
			wcex.hIconSm = 0; 

			atomRet = (ATOM) RegisterClassEx(&wcex);

			fRegisteredClass = true;

			//if (atomRet != 0)
			//	printf("\n\nRegistered class.\n");
		}

		return atomRet;
	}


	static LRESULT CALLBACK WndProc(HWND hwnd, UINT wMsg,
			WPARAM wParam, LPARAM lParam)
	{
		switch (wMsg)
		{
		case WM_PAINT:
			CImViewW32::StaticOnPaint(hwnd);
			break;

		case WM_DESTROY:
			CImViewW32::StaticOnDestroy(hwnd);
			break;

		case WM_CREATE:
			// lParam should be a non-NULL pointer to a CreateStruct.  The first
			// member of the struct is a pointer to the CImViewW32 class that
			// created this window.  We should save that pointer here.
			assert(lParam != 0);
			assert(*((long *) lParam) != 0);
			SetWindowLong(hwnd, GWL_USERDATA, *((long *) lParam));
			break;
		
		default:
			return (DefWindowProc(hwnd, wMsg, wParam, lParam));
		}

		return (0);
	}


	static void StaticOnPaint(HWND hwnd)
	{
		assert(hwnd != 0);

		CImViewW32 *pimagewnd = (CImViewW32 *)
				(void *) GetWindowLong(hwnd, GWL_USERDATA);
		assert(pimagewnd != 0);
		assert(pimagewnd->m_hwnd == hwnd);

		pimagewnd->OnPaint();
	}

	static void StaticOnDestroy(HWND hwnd)
	{
		assert(hwnd != 0);

		CImViewW32 *pimagewnd = (CImViewW32 *)
				(void *) GetWindowLong(hwnd, GWL_USERDATA);
		assert(pimagewnd != 0);
		assert(pimagewnd->m_hwnd == hwnd);

		pimagewnd->OnDestroy();
	}

	static unsigned __stdcall StaticThreadMain(void *pvCmdWnd)
	{
		CImViewW32 *pimagewnd = (CImViewW32 *) pvCmdWnd;
		assert(pimagewnd != 0);
		return pimagewnd->ThreadMain();
	}

	static DWORD WINAPI StaticThreadMain2(void *pvCmdWnd)
	{
		CImViewW32 *pimagewnd = (CImViewW32 *) pvCmdWnd;
		assert(pimagewnd != 0);
		return pimagewnd->ThreadMain();
	}



	void OnPaint(void)
	{
		PAINTSTRUCT ps;
		HDC hdc = BeginPaint(m_hwnd, &ps);
		// UNDONE:  If no image, fill with a solid color?
		if (m_pbData != 0)
			MyDraw(hdc);
		else
		{
			RECT rect;
			GetClientRect(m_hwnd, &rect);
			HBRUSH blackbrush = CreateSolidBrush(RGB(0,0,0));
			FillRect(hdc, &rect, blackbrush);
		}
		EndPaint(m_hwnd, &ps);
	}

	void OnDestroy(void)
	{
		m_hwnd = 0;
		KillThread(0);
	}



	unsigned __stdcall ThreadMain(void)
	{
		m_fThreadRunning = true;

		/*assert(m_pimage != 0);*/
		Create(m_szWindowName);

		SetEvent(m_hEventInitialized);

		// Message loop
		MSG msg;
		while ((::GetMessage(&msg, 0, 0, 0) == 1) && (!m_fKillThread))
		{
			if((msg.message==WM_PAINT)&&(msg.hwnd==0)&&(m_hwnd!=0))
			{				
				HDC hdc = GetDC(m_hwnd);
				if (m_pbData != 0)
					MyDraw(hdc);
				ReleaseDC(m_hwnd,hdc);
			}
			::TranslateMessage(&msg);
			::DispatchMessage(&msg);
			// Sleep(0);
		}

	#ifdef USE_CMD_WND_DEBUG_STRINGS
		OutputDebugString("CImViewW32 thread exiting.\n");
	#endif // USE_CMD_WND_DEBUG_STRINGS

		// Destroy the window before destroying the thread.
		if (m_hwnd != 0)
		{
			DestroyWindow(m_hwnd);
			m_hwnd = 0;
		}

		m_fThreadRunning = false;

		return 0;
	}

private:
	HANDLE m_hEventInitialized;
	HANDLE m_hThread;
	DWORD m_dwThreadID;
	HWND m_hwnd;
	//CVisImageBase *m_pimage;
	bool m_fKillThread;
	bool m_fThreadRunning;
	static char s_szAppName[MAX_PATH];

	void MyDraw(HDC hDC)
	{
		
		BITMAPINFOHEADER* dibhdr = (BITMAPINFOHEADER*)m_DIBHeader;						
		RECT viewRect;
		
		GetClientRect(m_hwnd, &viewRect);
		SetStretchBltMode(hDC, COLORONCOLOR);

		dibhdr->biWidth = m_iStride/m_iChannels;
		dibhdr->biHeight = m_iFlip*m_iHeight;

		StretchDIBits(hDC, 0, 0, viewRect.right, viewRect.bottom, 
				0, 0, m_iWidth, m_iHeight, m_pbData, (BITMAPINFO*)dibhdr, DIB_RGB_COLORS, SRCCOPY);	
	}

	bool Create(const char *szTitle = 0)
	{
		bool fRet = false;

		/*if ((m_pimage != 0) && (m_pimage->IsValid()))
		{
			if (szTitle == 0)
				szTitle = m_pimage->Name();
			if (szTitle == 0)*/
				//szTitle = "Untitled Image";

			// Find the desired window dimensions
			//RECT rectT = m_pimage->Rect();
			RECT rectT;
			rectT.top = 0;
			rectT.bottom = m_iHeight-1;
			rectT.left = 0;
			rectT.right = m_iWidth-1;
			AdjustWindowRectEx(&rectT, WS_OVERLAPPEDWINDOW, false, 0);
			int dx = rectT.right - rectT.left;
			int dy = rectT.bottom - rectT.top;

			m_hwnd = CreateWindow(s_szAppName, szTitle, WS_OVERLAPPEDWINDOW,
					CW_USEDEFAULT, 0, dx, dy, 0, 0, ::GetModuleHandle(0),
					(void *) this);

			if (m_hwnd != 0)
			{ 
				// LATER:  Add an "activate window" flag?
				ShowWindow(m_hwnd, SW_SHOWNA);
				UpdateWindow(m_hwnd);
				fRet = true;
			}
		/*}*/

		return fRet;
	}
	// Private copy constructor and assignment operator to prevent the compiler
	// from providing default implementations.
	CImViewW32(const CImViewW32& refimagewnd);
	CImViewW32& operator=(const CImViewW32& refimagewnd);
};


char CImViewW32::s_szAppName[MAX_PATH];


#endif // IMVIEWW32_H