
#include <stdio.h> //for _sprintf

//classes for image display
#include "imvieww32.h"

//classes for timings
#include "timer.h"

//classes for avifile reading
#include "avifile.h"


//GDI+ is used for image file reading
#include <gdiplus.h>
using namespace Gdiplus;
#pragma comment(lib, "gdiplus.lib")


INT main()
{
   const int maxchar = 128;
   const char caption[] = "Framework Test";
   char strbuffer[maxchar];

   GdiplusStartupInput gdiplusStartupInput;
   ULONG_PTR gdiplusToken;
   if(GdiplusStartup(&gdiplusToken, &gdiplusStartupInput, NULL) != Ok)
   {
	   MessageBox(NULL,"Cannot initialize GDI+ \n", caption, MB_OK);
	   return -1;
   }
   else
   {

	   MessageBox(NULL,"Starting single image processing",caption,MB_OK);
	   Bitmap bitmap(L"sunflower.bmp");
	   if(bitmap.GetLastStatus() != Ok)
	   {
		   MessageBox(NULL,"Error Reading File",caption, MB_OK);
		   GdiplusShutdown(gdiplusToken);
		   return -1;

	   }
	   
	   //getting image size
	   int bmpwidth = bitmap.GetWidth();
	   int bmpheight = bitmap.GetHeight();
	   int bmpbpp;
	   int bmpstride;
	   //getting image pixel format
	   PixelFormat bmppixfmt = bitmap.GetPixelFormat();
	   
	   switch(bmppixfmt)
	   {
	   case PixelFormat32bppRGB: 
		   bmpbpp = 32;
		   break;
	   case PixelFormat24bppRGB:
		   bmpbpp = 24;
		   break;
	   case PixelFormat8bppIndexed:
		   bmpbpp = 8;
		   break;
	   default:
		   bmpbpp = 0;
		   break;
	   }
	   Rect bmprect(0, 0, bmpwidth, bmpheight);

	   //poiter for raw image data7
	   BitmapData bitmapData;
	   bitmap.LockBits(&bmprect,ImageLockModeRead,bmppixfmt,&bitmapData);
	   byte* bmppixels = (byte*)bitmapData.Scan0;
	   bmpstride = bitmapData.Stride;	//careful - stride can be negative
	   int absstride = abs(bmpstride);  // now this is positive

	   CTimer t; //timer for benchmarking algorithms
	   t.StartClock();
	   //processing the image - compute the negative image
	   byte *bmpprocimg = (byte*)malloc(bmpstride*bmpheight);
	   for(int bmplines = 0; bmplines < bmpheight; bmplines ++)
		   for(int bmpbyteidx = 0; bmpbyteidx < absstride; bmpbyteidx++)
			   bmpprocimg[bmpstride*bmplines+bmpbyteidx] = (byte)(255-bmppixels[bmpstride*bmplines+bmpbyteidx]);
	   float time;
	   t.ReadClock(&time);
	   //release raw data
	   bitmap.UnlockBits(&bitmapData);

  	   //creating viewers for original and processed image
	   CImViewW32 wnd1("original",bmpwidth,bmpheight,bmpbpp);
	   CImViewW32 wnd2("processed",bmpwidth,bmpheight,bmpbpp);

	   //positionning viewers in desktop
	   wnd1.SetPos(50,0);
	   wnd2.SetPos(bmpwidth+100,0);

	   //setting data to image viewers
	   wnd1.SetData(bmppixels, bmpstride);  //use the signed stride - if negative, bitmap is bottom-up
	   wnd1.Refresh();	   
	   wnd2.SetData(bmpprocimg, bmpstride);
	   wnd2.Refresh();
	   _snprintf(strbuffer, maxchar, "Image processing took %f seconds.", time);

	   MessageBox(NULL,strbuffer, caption, MB_OK );

	   
	   MessageBox(NULL,"Done single image processing. Starting sequence processing",caption, MB_OK);
	   CAviFile avi;
	   LPBITMAPINFO videoformat;
	   // Open AVI...
	   avi.Open("highway.avi");
	   if(avi.GetVideoStreamCount() == 0)
			return -1;
	   videoformat = avi.GetVideoFormat(0);

	   // Start video retrieve... (this file only has the one stream)
	   avi.StartVideoRetrieve(0);

	   
	   LPBITMAPINFOHEADER pAvi;
	   int frame = 0;
	   avi.GetVideoFrame(0, frame++, &pAvi);

	   int avitopdown = 1;
	   if(pAvi->biHeight > 0)
			avitopdown = -1;

	   int aviheight = abs(pAvi->biHeight); 
	   int aviwidth = pAvi->biWidth;
	   int avibpp = pAvi->biBitCount;
	   int avistride = aviwidth*avibpp/8;
	   // pointer for image data
	   LPBYTE avipixels = (LPBYTE)pAvi + pAvi->biSize;

	   //allocate memory for processed image
	   byte *aviprocimg = (byte*)malloc(avistride*aviheight);

	   //creating viewers for original and processed image
	   CImViewW32 wnd3("original",aviwidth,aviheight,avibpp);
	   CImViewW32 wnd4("processed",aviwidth,aviheight,avibpp);

	   //positioning viewers in desktop
	   wnd3.SetPos(50,bmpheight+100);
	   wnd4.SetPos(aviwidth+100,bmpheight+100);


	   t.StartTimer();
	   //process all images of the sequence
	   while(pAvi != NULL)
	   {
		   t.MarkSectionStart();
		   for(int avilines = 0; avilines < aviheight; avilines ++)
				for(int avibyteidx = 0; avibyteidx < avistride; avibyteidx++)
					aviprocimg[avistride*avilines+avibyteidx] = (byte)(255-avipixels[avistride*avilines+avibyteidx]);
		   t.MarkSectionEnd();
		   //setting data to image viewers
		   wnd3.SetData(avipixels, avistride*avitopdown);
		   wnd3.Refresh();	   
		   wnd4.SetData(aviprocimg, avistride*avitopdown);
		   wnd4.Refresh();
		   avi.GetVideoFrame(0, frame++, &pAvi);
		   Sleep(100); //give some time to operating system
	   }
	   t.GetAvgProcessingTime(&time);
	   _snprintf(strbuffer, maxchar, "Average Processing Time %f sec.\n",time);
	   MessageBox(NULL,strbuffer, caption, MB_OK );

	   //closing avi functions
	   avi.EndVideoRetrieve(0);
	   avi.Release();

	   MessageBox(NULL,"Done sequence processing. Exiting...",caption, MB_OK);

	   //releasing allocated memory
	   if(bmpprocimg) free(bmpprocimg);
	   if(aviprocimg) free(aviprocimg);
   }
   
   GdiplusShutdown(gdiplusToken);
   return 0;
}


